"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.initializeLanguageServiceInstance = void 0;
const bs_logger_1 = require("bs-logger");
const fs_1 = require("fs");
const path_1 = require("path");
const memoize = require("lodash.memoize");
const mkdirp = require("mkdirp");
const constants_1 = require("../constants");
const messages_1 = require("../utils/messages");
const json_1 = require("../utils/json");
const sha1_1 = require("../utils/sha1");
function doTypeChecking(configs, diagnosedFiles, fileName, service, logger) {
    if (configs.shouldReportDiagnostics(fileName)) {
        const diagnostics = service.getSemanticDiagnostics(fileName).concat(service.getSyntacticDiagnostics(fileName));
        diagnosedFiles.push(fileName);
        configs.raiseDiagnostics(diagnostics, fileName, logger);
    }
}
exports.initializeLanguageServiceInstance = (configs, logger) => {
    logger.debug('initializeLanguageServiceInstance(): create typescript compiler');
    const ts = configs.compilerModule;
    const cwd = configs.cwd;
    const cacheDir = configs.tsCacheDir;
    const { options, fileNames } = configs.parsedTsConfig;
    const diagnosedFiles = [];
    const serviceHostTraceCtx = {
        namespace: 'ts:serviceHost',
        call: null,
        [bs_logger_1.LogContexts.logLevel]: bs_logger_1.LogLevels.trace,
    };
    const memoryCache = {
        files: new Map(),
        resolvedModules: new Map(),
    };
    let tsResolvedModulesCachePath;
    if (cacheDir) {
        mkdirp.sync(cacheDir);
        tsResolvedModulesCachePath = path_1.join(cacheDir, sha1_1.sha1('ts-jest-resolved-modules', '\x00'));
        try {
            const cachedTSResolvedModules = fs_1.readFileSync(tsResolvedModulesCachePath, 'utf-8');
            memoryCache.resolvedModules = new Map(json_1.parse(cachedTSResolvedModules));
        }
        catch (e) { }
    }
    configs.parsedTsConfig.fileNames
        .filter((fileName) => !configs.isTestFile(fileName))
        .forEach((fileName) => {
        memoryCache.files.set(fileName, {
            version: 0,
        });
    });
    function isFileInCache(fileName) {
        return memoryCache.files.has(fileName) && memoryCache.files.get(fileName).version !== 0;
    }
    const cacheReadFile = logger.wrap(serviceHostTraceCtx, 'readFile', memoize(ts.sys.readFile));
    const moduleResolutionHost = {
        fileExists: memoize(ts.sys.fileExists),
        readFile: cacheReadFile,
        directoryExists: memoize(ts.sys.directoryExists),
        getCurrentDirectory: () => cwd,
        realpath: ts.sys.realpath && memoize(ts.sys.realpath),
        getDirectories: memoize(ts.sys.getDirectories),
    };
    const moduleResolutionCache = ts.createModuleResolutionCache(cwd, (x) => x, options);
    function resolveModuleNames(moduleNames, containingFile) {
        var _a;
        const normalizedContainingFile = path_1.normalize(containingFile);
        const currentResolvedModules = (_a = memoryCache.resolvedModules.get(normalizedContainingFile)) !== null && _a !== void 0 ? _a : [];
        return moduleNames.map((moduleName) => {
            const resolveModuleName = ts.resolveModuleName(moduleName, containingFile, options, moduleResolutionHost, moduleResolutionCache);
            const resolvedModule = resolveModuleName.resolvedModule;
            if (configs.isTestFile(normalizedContainingFile) && resolvedModule) {
                const normalizedResolvedFileName = path_1.normalize(resolvedModule.resolvedFileName);
                if (!currentResolvedModules.includes(normalizedResolvedFileName)) {
                    currentResolvedModules.push(normalizedResolvedFileName);
                    memoryCache.resolvedModules.set(normalizedContainingFile, currentResolvedModules);
                }
            }
            return resolvedModule;
        });
    }
    let projectVersion = 1;
    const updateMemoryCache = (contents, fileName) => {
        logger.debug({ fileName }, 'updateMemoryCache(): update memory cache for language service');
        let shouldIncrementProjectVersion = false;
        const hit = isFileInCache(fileName);
        if (!hit) {
            memoryCache.files.set(fileName, {
                text: contents,
                version: 1,
            });
            shouldIncrementProjectVersion = true;
        }
        else {
            const previousContents = memoryCache.files.get(fileName).text;
            if (previousContents !== contents) {
                memoryCache.files.set(fileName, {
                    text: contents,
                    version: memoryCache.files.get(fileName).version + 1,
                });
                if (hit)
                    shouldIncrementProjectVersion = true;
            }
            if (!fileNames.includes(fileName)) {
                shouldIncrementProjectVersion = true;
            }
        }
        if (shouldIncrementProjectVersion)
            projectVersion++;
    };
    const serviceHost = {
        getProjectVersion: () => String(projectVersion),
        getScriptFileNames: () => [...memoryCache.files.keys()],
        getScriptVersion: (fileName) => {
            var _a;
            const normalizedFileName = path_1.normalize(fileName);
            const version = (_a = memoryCache.files.get(normalizedFileName)) === null || _a === void 0 ? void 0 : _a.version;
            return version === undefined ? undefined : String(version);
        },
        getScriptSnapshot(fileName) {
            var _a;
            const normalizedFileName = path_1.normalize(fileName);
            const hit = isFileInCache(normalizedFileName);
            logger.trace({ normalizedFileName, cacheHit: hit }, 'getScriptSnapshot():', 'cache', hit ? 'hit' : 'miss');
            if (!hit) {
                memoryCache.files.set(normalizedFileName, {
                    text: cacheReadFile(normalizedFileName),
                    version: 1,
                });
            }
            const contents = (_a = memoryCache.files.get(normalizedFileName)) === null || _a === void 0 ? void 0 : _a.text;
            if (contents === undefined)
                return;
            return ts.ScriptSnapshot.fromString(contents);
        },
        fileExists: memoize(ts.sys.fileExists),
        readFile: cacheReadFile,
        readDirectory: memoize(ts.sys.readDirectory),
        getDirectories: memoize(ts.sys.getDirectories),
        directoryExists: memoize(ts.sys.directoryExists),
        realpath: ts.sys.realpath && memoize(ts.sys.realpath),
        getNewLine: () => constants_1.LINE_FEED,
        getCurrentDirectory: () => cwd,
        getCompilationSettings: () => options,
        getDefaultLibFileName: () => ts.getDefaultLibFilePath(options),
        getCustomTransformers: () => configs.customTransformers,
        resolveModuleNames,
    };
    logger.debug('initializeLanguageServiceInstance(): creating language service');
    const service = ts.createLanguageService(serviceHost, ts.createDocumentRegistry());
    return {
        compileFn: (code, fileName) => {
            var _a;
            logger.debug({ fileName }, 'compileFn(): compiling using language service');
            updateMemoryCache(code, fileName);
            const output = service.getEmitOutput(fileName);
            if (tsResolvedModulesCachePath) {
                void (() => __awaiter(void 0, void 0, void 0, function* () {
                    yield fs_1.writeFile(tsResolvedModulesCachePath, json_1.stringify([...memoryCache.resolvedModules]), () => { });
                }))();
            }
            if (!diagnosedFiles.includes(fileName)) {
                logger.debug({ fileName }, 'compileFn(): computing diagnostics using language service');
                doTypeChecking(configs, diagnosedFiles, fileName, service, logger);
            }
            if (!configs.isTestFile(fileName)) {
                for (const [testFileName, resolvedModules] of memoryCache.resolvedModules.entries()) {
                    if (resolvedModules.includes(fileName) &&
                        !diagnosedFiles.includes(testFileName) &&
                        fs_1.existsSync(testFileName)) {
                        const testFileContent = (_a = memoryCache.files.get(testFileName)) === null || _a === void 0 ? void 0 : _a.text;
                        if (!testFileContent) {
                            updateMemoryCache(cacheReadFile(testFileName), testFileName);
                        }
                        logger.debug({ testFileName }, 'compileFn(): computing diagnostics using language service for test file which uses the module');
                        doTypeChecking(configs, diagnosedFiles, testFileName, service, logger);
                    }
                }
            }
            if (output.emitSkipped) {
                throw new TypeError(`${path_1.relative(cwd, fileName)}: Emit skipped for language service`);
            }
            if (!output.outputFiles.length) {
                throw new TypeError(messages_1.interpolate("Unable to require `.d.ts` file for file: {{file}}.\nThis is usually the result of a faulty configuration or import. Make sure there is a `.js`, `.json` or another executable extension available alongside `{{file}}`.", {
                    file: path_1.basename(fileName),
                }));
            }
            return [output.outputFiles[1].text, output.outputFiles[0].text];
        },
        program: service.getProgram(),
    };
};
